"""
expedite
Copyright (C) 2024 Akashdeep Dhar

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.

Any Red Hat trademarks that are incorporated in the codebase or documentation
are not subject to the GNU General Public License and may only be utilized or
replicated with the express permission of Red Hat, Inc.
"""


import sys
from asyncio import ensure_future, get_event_loop
from json import loads

from websockets import connect
from websockets.exceptions import ConnectionClosed

from expedite.client.conn import (
    collect_confirmation,
    collect_connection_from_pairness,
    collect_contents,
    collect_digest_checks,
    collect_dropping_summon,
    collect_metadata,
    collect_permission_to_join,
    collect_separation_from_mistaken_password,
    deliver_confirmation,
    deliver_connection_to_server,
    deliver_contents,
    deliver_digest_checks,
    deliver_dropping_summon,
    deliver_metadata,
    deliver_separation_from_mistaken_password,
    deliver_suspension_from_expiry,
    facade_exit,
)
from expedite.config import standard


async def oper():
    try:
        async with connect(standard.client_addr) as sock:
            get_event_loop().call_later(standard.client_time, lambda: ensure_future(deliver_suspension_from_expiry(sock)))
            await deliver_connection_to_server(sock)
            async for mesgcont in sock:
                if isinstance(mesgcont, str):
                    mesgdict = loads(mesgcont)
                    # If the data received is of STRING type
                    if standard.client_plan in ["SEND", "RECV"]:
                        # If the purpose of the client is either DELIVERING or COLLECTING
                        if mesgdict["call"] == "okay":
                            await collect_permission_to_join(mesgdict["iden"])
                        elif mesgdict["call"] in ["awry", "lone"]:
                            await facade_exit(sock, False, mesgdict["call"])
                    if standard.client_plan == "SEND":
                        # If the purpose of the client is DELIVERING
                        if mesgdict["call"] == "note":
                            await collect_connection_from_pairness(mesgdict["part"])
                            await deliver_metadata(sock)
                        elif mesgdict["call"] == "conf":
                            complete = await collect_confirmation(mesgdict["data"])
                            await facade_exit(sock, complete, "done" if complete else "dprt")
                        elif mesgdict["call"] == "flub":
                            await collect_separation_from_mistaken_password()
                            await facade_exit(sock, False, "flub")
                        elif mesgdict["call"] == "drop":
                            await collect_dropping_summon()
                            await deliver_contents(sock)
                            await deliver_digest_checks(sock)
                    else:
                        # If the purpose of the client is COLLECTING
                        if mesgdict["call"] == "note":
                            await collect_connection_from_pairness(mesgdict["part"])
                        elif mesgdict["call"] == "hash":
                            await collect_digest_checks()
                            complete = await deliver_confirmation(sock, mesgdict["data"])
                            await facade_exit(sock, complete, "done" if complete else "dprt")
                else:
                    # If the data received is of BYTES type
                    if standard.client_plan == "RECV":
                        # If the purpose of the client is COLLECTING
                        if not standard.client_metadone:
                            if await collect_metadata(mesgcont):
                                await deliver_dropping_summon(sock)
                            else:
                                await deliver_separation_from_mistaken_password(sock)
                                await facade_exit(sock, False, "flub")
                        else:
                            await collect_contents(sock, mesgcont)
            sys.exit(standard.client_exit)
    except ConnectionClosed:
        await facade_exit(sock, False, "dprt")
        sys.exit(standard.client_exit)



"""
expedite
Copyright (C) 2024 Akashdeep Dhar

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.

Any Red Hat trademarks that are incorporated in the codebase or documentation
are not subject to the GNU General Public License and may only be utilized or
replicated with the express permission of Red Hat, Inc.
"""


from json import dumps
from uuid import uuid4

from websockets.legacy.server import WebSocketServerProtocol

from expedite.config import standard
from expedite.view import failure, general, success, warning


async def exchange_insert(sock: WebSocketServerProtocol, plan: str = standard.client_plan, scan: str = standard.client_endo, time: int = standard.client_time) -> str | bool:
    if sock not in standard.connection_dict:
        if plan in ["SEND", "RECV"]:
            identity = uuid4().hex[0:8].upper()
            standard.connection_dict[sock] = {
                "iden": identity,
                "plan": plan,
                "scan": scan if scan != "" else None,
                "time": time,
                "pair": False,
                "ptid": "",
                "ptsc": None
            }
            if plan == "SEND":
                warning(f"{identity} joined with the intention of delivering.")
            elif plan == "RECV":
                warning(f"{identity} joined with the intention of collecting.")
            if scan == "":
                general(f"{identity} is waiting for client for {time} seconds.")
            else:
                general(f"{identity} is looking for {scan} for {time} seconds.")
            await sock.send(dumps({"call": "okay", "iden": identity}))
            return identity
        else:
            return False
    else:
        return False


async def exchange_remove(sock: WebSocketServerProtocol) -> bool:
    if sock in standard.connection_dict:
        if standard.connection_dict[sock]["ptsc"] in standard.connection_dict and standard.connection_dict[sock]["ptsc"].state == 1:
            warning(f"{standard.connection_dict[sock]["ptid"]} left.")
            await standard.connection_dict[sock]["ptsc"].close(code=1000)
            standard.connection_dict.pop(standard.connection_dict[sock]["ptsc"])
        warning(f"{standard.connection_dict[sock]["iden"]} left.")
        await sock.close(code=1000)
        standard.connection_dict.pop(sock)
        return True
    else:
        return False


async def exchange_inform(sock: WebSocketServerProtocol, plan: str = standard.client_plan, scan: str = standard.client_endo, iden: str = standard.client_iden) -> int:
    for indx in standard.connection_dict:
        if standard.connection_dict[indx]["iden"] == scan:
            if not standard.connection_dict[indx]["pair"]:
                if plan != standard.connection_dict[indx]["plan"]:
                    success(f"{iden} and {scan} are positively paired.")
                    await indx.send(dumps({"call": "note", "part": iden}))
                    await sock.send(dumps({"call": "note", "part": scan}))
                    standard.connection_dict[indx]["pair"], standard.connection_dict[sock]["pair"] = True, True
                    standard.connection_dict[indx]["ptid"], standard.connection_dict[sock]["ptid"] = iden, scan
                    standard.connection_dict[indx]["ptid"], standard.connection_dict[sock]["ptid"] = iden, scan
                    standard.connection_dict[indx]["ptsc"], standard.connection_dict[sock]["ptsc"] = sock, indx
                    return 0
                else:
                    failure(f"{iden} and {scan} are negatively paired.")
                    await indx.send(dumps({"call": "awry", "part": iden}))
                    await sock.send(dumps({"call": "awry", "part": scan}))
                    return 1
            else:
                failure(f"{iden} and {scan} cannot pair as {scan} is already paired.")
                await sock.send(dumps({"call": "lone", "part": scan}))
                return 2
    return 3


async def exchange_json(sock: WebSocketServerProtocol, note: str = "", data: str = "") -> bool:
    general(standard.server_note[note].format(sj=standard.connection_dict[sock]["iden"], oj=standard.connection_dict[sock]["ptid"]))
    if sock in standard.connection_dict:
        if standard.connection_dict[sock]["ptsc"] in standard.connection_dict and standard.connection_dict[sock]["ptsc"].state == 1:
            await standard.connection_dict[sock]["ptsc"].send(data)
            return True
        else:
            return False
    else:
        return False


async def exchange_byte(sock: WebSocketServerProtocol, pack: bytes = b"") -> bool:
    if sock in standard.connection_dict:
        if standard.connection_dict[sock]["ptsc"] in standard.connection_dict and standard.connection_dict[sock]["ptsc"].state == 1:
            await standard.connection_dict[sock]["ptsc"].send(pack)
            return True
        else:
            general(f"{standard.connection_dict[sock]["iden"]} could not deliver contents to {standard.connection_dict[sock]["ptid"]} as {standard.connection_dict[sock]["ptid"]} is no longer connected.")
            return False
    else:
        general("Attempting for aborting connection.")
        return False
